
\begin{code}
module ParsePackets (parse_packets) where

import List
import Word
import Maybe
import Char
import Bits

import Packet

parse_packets :: String -> [Packet]
parse_packets = parse_packets' . map ord

parse_packets' :: [Int] -> [Packet]
parse_packets' [] = []
parse_packets' (i:is)
  | i .&. 2^7 == 0 = error "First bit of a packet must be 1!"
  | otherwise      = parse_packets'' $ get_packet_details (i `mod` 2^6) is
    where get_packet_details = if i .&. 2^6 == 0 then get_old_packet_details
                                                 else get_new_packet_details

parse_packets'' :: ([Int], Tag, [Int]) -> [Packet]
parse_packets'' (this, t, rest) = parse_packet t this:parse_packets' rest

splitAtExactly :: Int -> [a] -> ([a], [a])
splitAtExactly 0 xs = ([], xs)
splitAtExactly (n+1) (x:xs) = (x:ys, zs)
    where (ys, zs) = splitAtExactly n xs
splitAtExactly _ _ = error "splitAtExactly"

get_new_packet_details :: Tag -> [Int] -> ([Int], Tag, [Int])
get_new_packet_details t is = (is1, t, is2)
    where (is1, is2) = get_new_packet is

get_new_packet :: [Int] -> ([Int], [Int])
get_new_packet (i:is)
    | i < 192 = splitAt i is
    | i >= 224 && i < 255 = let log_len = fromIntegral i .&. 0x1F :: Word32
                                len = shiftL 1 (fromIntegral log_len) :: Word32
                                (is1, is2) = genericSplitAt len is
                                (is3, is4) = get_new_packet is2
                            in (is1 ++ is3, is4)
get_new_packet (i1:i2:is)
    | i1 < 224 = splitAt ((i1 - 192) `shiftL` 8 + i2 + 192) is
get_new_packet (255:i1:i2:i3:i4:is) = splitAt len is
    where len = i1 `shiftL` 24 + i2 `shiftL` 16 + i3 `shiftL` 8  + i4
get_new_packet _ = error "get_new_packet ran out of input reading length"

get_old_packet_details :: Tag -> [Int] -> ([Int], Tag, [Int])
get_old_packet_details t is = (is1, content_tag, is2)
    where content_tag = t `shiftR` 2
          length_type = t .&. 3
          (len, is') = get_old_length length_type is
          (is1, is2) = splitAt len is'

parse_packet :: Tag -> [Int] -> Packet
parse_packet t = fromMaybe (parse_other t) $ lookup t packet_parsers

-- XXX This list could be generated by template Haskell?
packet_parsers :: [(Tag, ([Int] -> Packet))]
packet_parsers = [(3, parse_t3_sk_esk), (8, parse_t8_cd), (9, parse_t9_sed),
                  (11, parse_t11_ld)]

-- This could be nicer recursively?
get_old_length :: Tag -> [Int] -> (Length, [Int])
get_old_length 0 (i:is) = (i, is)
get_old_length 1 (i1:i2:is) = (i1 `shiftL` 8 + i2, is)
get_old_length 2 (i1:i2:i3:i4:is) = (i1 `shiftL` 24 + i2 `shiftL` 16
                               + i3 `shiftL` 8  + i4, is)
get_old_length 3 is = (length is, is)
get_old_length _ _ = error "Ran out of input reading length (or length type > 3, but that can't happen...)"

-- Is this right?
parse_t3_sk_esk :: [Int] -> Packet
parse_t3_sk_esk (4:alg_id:is) = T3_SK_ESK (id_to_sym_alg alg_id) s2k esk
    where (s2k, is') = get_s2k is
          esk = if null is' then Nothing else Just is'
parse_t3_sk_esk _ = error "parse_t3_sk_esk"

get_s2k :: [Int] -> (S2K, [Int])
get_s2k (0:hash:is) = (SimpleS2K (id_to_hash hash), is)
get_s2k (1:hash:s7:s6:s5:s4:s3:s2:s1:s0:is) = (SaltedS2K (id_to_hash hash) salt, is)
    where salt = make_salt [s7, s6, s5, s4, s3, s2, s1, s0]
get_s2k (3:hash:s7:s6:s5:s4:s3:s2:s1:s0:count:is) = (IteratedSaltedS2K (id_to_hash hash) salt count, is)
    where salt = make_salt [s7, s6, s5, s4, s3, s2, s1, s0]
get_s2k _ = error "get_s2k"

make_salt :: [Int] -> Salt
-- make_salt = foldl (\t s -> 256*t + s) 0 . map fromIntegral
make_salt = map chr

-- XXX incomplete
id_to_hash :: AlgID -> Hash
id_to_hash alg_id = fromJust $ lookup alg_id (zip [1..] [MD5, SHA1])

-- XXX incomplete
id_to_sym_alg :: AlgID -> SymAlg
id_to_sym_alg alg_id = fromJust $ lookup alg_id [(3, CAST5)]

-- XXX incomplete
id_to_comp_alg :: AlgID -> CompAlg
id_to_comp_alg alg_id = fromJust $ lookup alg_id [(1, ZIP)]

parse_t8_cd :: [Int] -> Packet
parse_t8_cd (comp_alg_id:cs) = T8_CD (id_to_comp_alg comp_alg_id) cs
parse_t8_cd _ = error "parse_t8_cd"

parse_t9_sed :: [Int] -> Packet
parse_t9_sed = T9_SED

parse_t11_ld :: [Int] -> Packet
parse_t11_ld (m:filename_len:xs) = T11_LD mode filename' date literal_data
    where mode = get_mode m
          (filename, xs') = splitAt filename_len xs
          filename' = map chr filename
          date = foldl (\t i -> t * 256 + i) 0 $ map fromIntegral $ take 4 xs'
          literal_data = map chr $ drop 4 xs'

get_mode :: Int -> Mode
get_mode 98 = Binary
get_mode 116 = Text
get_mode 108 = Local
get_mode _ = error "Invalid mode"

parse_other :: Tag -> [Int] -> Packet
parse_other t is = Other t is
\end{code}

